import os
import kabuki
import numpy as np
import unittest
from nose.tools import raises
import pymc as pm
from utils import HNodeSimple, HNodeSimpleVar, sample_from_models, create_test_models
import pandas as pd

from utils import gen_func_df

class TestHierarchicalBreakDown(unittest.TestCase):
    """
    simple tests to see if hierarchical merthods do not raise and error
    """

    @classmethod
    def setUpClass(self):
    #def setUp(self):

        #load models
        self.models, self.params = create_test_models()

        #run models
        sample_from_models(self.models, n_iter=200)

    def runTest(self):
        pass

    def test_map(self):
        np.random.seed(123)
        for model in self.models:
            if model.is_group_model:
                model.approximate_map()
            else:
                model.map(runs=2)

    def test_dic_info(self):
        for model in self.models:
            model.dic_info()

    def test_print_stats(self):
        for model in self.models:
            model.print_stats()
            model.print_stats(subj=False)
            if model.is_group_model:
                model.print_stats(subj_idx=1)

    def test_load_db(self):
        new_models, params = create_test_models()
        for i, model in enumerate(new_models):
            print "sample model", i
            model.sample(100, dbname='unittest.db', db='pickle')
            model.load_db(dbname='unittest.db', db='pickle')
            model.gen_stats()
            os.remove('unittest.db')

    @unittest.skip("TODO")
    def test_init_from_existing_model(self):
        new_models = create_test_models()
        for (i_m, pre_model) in enumerate(self.models):
            new_models[i_m].init_from_existing_model(pre_model)

    def test_plot_posteriors(self):
        pass

    @unittest.skip("TODO")
    def test_subj_by_subj_map_init(self):
        models = create_test_models()
        for model in models:
            if model.is_group_model:
                model.subj_by_subj_map_init(runs=1)


class TestModelCreation(unittest.TestCase):

    @classmethod
    def setUpClass(self):
#    def setUp(self):
        self.n_subj = 3
        data, _ = kabuki.generate.gen_rand_data(gen_func_df, {'A':{'loc':0, 'scale':1}, 'B': {'loc':0, 'scale':1}},
                                                subjs=self.n_subj)
        data = pd.DataFrame(data)
        data['condition2'] = np.random.randint(2, size=len(data))
        self.data = data

    def runTest(self):
        pass

    def test_save_and_load_breakdown(self):
        m = HNodeSimple(self.data)
        m.sample(500, dbname='test.db', db='pickle')
        m.save('test.model')
        m_load = kabuki.utils.load('test.model')
        os.remove('test.db')
        os.remove('test.model')

    def test_simple_no_deps(self):
        m = HNodeSimple(self.data)
        n_nodes = 1 + self.n_subj*2 #mu_g + n_subj * (mu_subj + like)
        self.assertEqual(len(m.nodes_db), n_nodes)

    def test_simple_deps(self):
        m = HNodeSimple(self.data, depends_on={'mu': 'condition'})
        n_nodes = 2 * (1 + self.n_subj*2) #n_conds * (mu_g + n_subj * (mu_subj + like))
        self.assertEqual(len(m.nodes_db), n_nodes)

    @raises(AssertionError)
    def test_assertion_on_wrong_param_name(self):
        HNodeSimple(self.data, depends_on={'non_existant': 'condition'})

    def test_assertion_on_wrong_param_name_tofix(self):
        # does not catch if correct argument
        HNodeSimple(self.data, depends_on={'non_existant': 'condition', 'mu': 'condition'})

    def test_simplevar_partly_deps(self):
        m = HNodeSimpleVar(self.data, depends_on={'mu': 'condition'})
        n_nodes = 1 + 1 + 2 * (1 + self.n_subj*2) #mu_std + mu_tau + n_conds * (mu_g + n_subj * (mu_subj + like))
        self.assertEqual(len(m.nodes_db), n_nodes)

    def test_simplevar_deps(self):
        m = HNodeSimpleVar(self.data, depends_on={'mu': 'condition', 'mu_std':'condition'})
        #n_nodes = n_conds * (mu_std + mu_tau + mu_g + n_subj * (mu_subj + like))
        n_nodes = 2 * (1 + 1 + 1 + self.n_subj*2)
        self.assertEqual(len(m.nodes_db), n_nodes)

    def test_simplevar_double_deps_A(self):
        m = HNodeSimpleVar(self.data, depends_on={'mu': 'condition', 'mu_std':'condition2'})
        #n_nodes = 2*v_tau + 2*v_std + 2*v_g + 4*n_subj*(v_subj + like))
        n_nodes = 2 + 2 + 2 + 4 * self.n_subj * 2
        self.assertEqual(len(m.nodes_db), n_nodes)

    def test_simplevar_double_deps_B(self):
        m = HNodeSimpleVar(self.data, depends_on={'mu': ['condition', 'condition2'], 'mu_std':'condition2'})
        #n_nodes = 2*mu_tau + 2*mu_std + 4*mu_g + 4*n_subj*(mu_subj + like))
        n_nodes = 2 + 2 + 4 + 4 * self.n_subj * 2
        self.assertEqual(len(m.nodes_db), n_nodes)


class TestEstimation(unittest.TestCase):
    """
    simple tests to see if hierarchical methods do not raise and error
    """

    def test_map_approx(self):
        subjs = 40
        data, params_true = kabuki.generate.gen_rand_data(gen_func_df,
                                                          {'A':{'loc':0, 'scale':1}, 'B': {'loc':2, 'scale':1}},
                                                          subj_noise={'loc':.1}, size=1000, subjs=subjs)

        model = HNodeSimple(data, depends_on={'mu': 'condition'})

        model.approximate_map()

        counter = 0
        for condition, subj_params in params_true.iteritems():
            nodes = model.nodes_db[model.nodes_db['condition'] == condition]
            for idx, params in enumerate(subj_params):
                nodes_subj = nodes[nodes['subj_idx'] == idx]
                for param_name, value in params.iteritems():
                    if param_name != 'loc':
                        continue
                    node = nodes_subj[nodes_subj.knode_name == 'mu_subj']
                    assert len(node) == 1, "Only one node should have been left after slicing."
                    abs_error = np.abs(node.map[0] - value)
                    self.assertTrue(abs_error < .2)
                    counter += 1

        # looping for each condition (i.e. twice)
        self.assertEqual(counter, subjs*2)


class TestConcatenate(unittest.TestCase):
    def test_concat(self):
        n_subj = 5
        data, params = kabuki.generate.gen_rand_data(gen_func_df, {'A':{'loc':0, 'scale':1}, 'B': {'loc':0, 'scale':1}}, subjs=n_subj)
        data = pd.DataFrame(data)

        models = []
        for i in range(4):
            m = HNodeSimple(data)
            m.sample(50, burn=0, db='pickle', dbname='test_%d'%i)
            models.append(m)

        super_model = kabuki.utils.concat_models(models)
        stochs = super_model.get_stochastics()
        for stoch in stochs.node:
            self.assertEqual(len(stoch.trace[:]), 50*4)

        for i in range(4):
            os.remove('test_%d'%i)




from __future__ import division

from copy import deepcopy

import numpy as np
import pandas as pd


def _add_noise(params, check_valid_func=None, bounds=None, noise=.1, exclude_params=(), share_noise=()):
    """Add individual noise to each parameter.

        :Arguments:
            params : dict
                Parameters to use for data generation. should look like this
                * {'condition1': {'param1': value, 'param2': value2},
                   'condition2': {'param1': value3, 'param2': value4}}
        :Optional:
            check_valid_func : function <default lambda x: True>
                Function that takes as input the parameters as kwds
                and returns True if param values are admissable.
            bounds : dict <default={}>
                Dict containing parameter names and
                (lower, upper) value for valid parameter
                range.
            noise : float <default=.1>
                Standard deviation of random gaussian
                variable to add to each parameter.
            exclude_params : tuple <default=()>
                Do not add noise to these parameters.
            share_noise : tuple <default=()>
                parameters in share_noise will share the noise across conditions

        :Returns:
            params : dict
                parameters with noise added.

    """

    def normal_rand(mu, sigma):
        if sigma == 0:
            return mu
        else:
            return np.random.normal(loc=mu, scale=sigma)

    def sample_value(param, value):
        if np.isscalar(noise):
            param_noise = noise
        else:
            param_noise = noise.get(param, 0)

        if param in bounds:
            while True:
                sampled_value = normal_rand(value, param_noise)
                if sampled_value >= bounds[param][0] and sampled_value <= bounds[param][1]:
                    return sampled_value
        else:
            return normal_rand(value, param_noise)

    if bounds is None:
        bounds = {}

    if check_valid_func is None:
        check_valid_func = lambda **params: True


    # Sample parameters until accepted
    original_params = deepcopy(params)
    params_noise= deepcopy(params)
    while True:
        params = deepcopy(original_params)

        # sample params only if not excluded and make sure they are shaared across condition if necessary
        for (i_cond, (cond, cond_params)) in enumerate(original_params.iteritems()):
            if i_cond == 0:
                cond0 = cond
            for param, value in cond_params.iteritems():
                if param not in exclude_params:
                    if (i_cond == 0) or (param not in share_noise):
                        new_value = sample_value(param, value)
                        params[cond][param] = new_value
                        params_noise[cond][param] = new_value - value
                    else:
                        params[cond][param] += params_noise[cond0][param]

        # check if params are valid
        valid = True
        for cond_params in params.itervalues():
            valid = check_valid_func(**cond_params)
            if not valid:
                break

        # return params
        if valid:
            return params

def gen_rand_data(gen_func, params, size=50, subjs=1, subj_noise=.1, exclude_params=(), share_noise=(),
                  column_name='data', check_valid_func=None, bounds=None, seed=None, generate_data=True):
    """Generate a random dataset using a user-defined random function.

    :Arguments:
        gen_func : A python function taking size and **param kwargs and returning a dataframe
        params : dict
            Parameters to use for data generation. Two options possible:
                * {'param1': value, 'param2': value2}
                * {'condition1': {'param1': value, 'param2': value2},
                   'condition2': {'param1': value3, 'param2': value4}}
            In the second case, the dataset is generated with multiple conditions
            named after the key and will be sampled using the corresponding parameters.

    :Optional:
        size : int <default: 50>
            How many values to sample for each condition for each subject.
        subjs : int <default: 1>
            How many subjects to generate data from. Individual subject parameters
            will be normal distributed around the provided parameters with variance
            subj_noise if subjs > 1. If only one subject is simulated no noise is added.
        subj_noise : float or dictionary <default: .1>
            How much to perturb individual subj parameters.
            if float then each parameter will be sampled from a normal distribution with std of subj_noise.
            if dictionary then only parameters that are keys of subj_noise will be sampled, and the std of the sampling
            distribution will be the value associated with them.
        exclude_params : tuple <default ()>
            Do not add noise to these parameters.
        share_noise : tuple <default=()>
            parameters in share_noise will share the noise across conditions
        check_valid_func : function <default lambda x: True>
            Function that takes as input the parameters as kwds
            and returns True if param values are admissable.
        bounds : dict <default={}>
            Dict containing parameter names and
            (lower, upper) value for valid parameter
            range.
        column_name : str <default='data'>
            What to name the data column.

    :Returns:
        data : numpy structured array
            Will contain the columns 'subj_idx', 'condition' and 'data' which contains
            the random samples.
        final_params_set : dict mapping condition to list of individual subject parameters
            Tries to be smart and will return direct values if there is only 1 subject
            and no dict if there is only 1 condition.

    """

    #The value of generate_data affects the parameters generation
    #so in the mean time we fix that, by completely ignoring the argument.
    generate_data = True

    # Check if only dict of params was passed, i.e. no conditions
    if not isinstance(params[params.keys()[0]], dict):
        params = {'none': params}
    if seed is not None:
        np.random.seed(seed)

    final_params_set = {}
    for condition in params.iterkeys():
        final_params_set[condition] = []

    data = []
    for subj_idx in range(subjs):
        #if it is a group model add noise to the parameters
        if subjs > 1:
            # Sample subject parameters from a normal around the specified parameters
            subj_params = _add_noise(params, noise=subj_noise, share_noise=share_noise,
                                     check_valid_func=check_valid_func,
                                     bounds=bounds,
                                     exclude_params=exclude_params)
        else:
            subj_params = params.copy()

        #sample for each condition
        for condition, params_cur in subj_params.iteritems():
            final_params_set[condition].append(params_cur)
            if generate_data:
                samples_from_dist = gen_func(size=size, **params_cur)
            else:
                samples_from_dist = np.empty(size,dtype=np.float)
                samples_from_dist[:] = np.nan

            samples_from_dist = pd.DataFrame(samples_from_dist)
            samples_from_dist['subj_idx'] = subj_idx
            samples_from_dist['condition'] = condition
            data.append(samples_from_dist)

    # Remove list around final_params_set if there is only 1 subject
    if subjs == 1:
        for key, val in final_params_set.iteritems():
            final_params_set[key] = val[0]

    # Remove dict around final_params_set if there is only 1 condition
    if len(final_params_set) == 1:
        final_params_set = final_params_set[final_params_set.keys()[0]]

    return pd.concat(data, ignore_index=True), final_params_set



from hierarchical import *

import utils
import analyze
import step_methods as steps
import generate

__version__ = '0.5.2.beta.dev'

try:
    from IPython.core.debugger import Tracer; debug_here = Tracer()
except:
    try:
        from IPython.Debugger import Tracer; debug_here = Tracer()
    except:
        def debug_here(): pass

try:
    from collections import OrderedDict
except:
    from OrderedDict import OrderedDict



from distutils.core import setup

setup(
    name="kabuki",
    version="0.5.2.beta",
    author="Thomas V. Wiecki, Imri Sofer",
    author_email="thomas.wiecki@gmail.com",
    url="http://github.com/hddm-devs/kabuki",
    packages=["kabuki"],
    description="kabuki is a python toolbox that allows easy creation of hierarchical bayesian models for the cognitive sciences.",
    install_requires=['NumPy >= 1.6.0', 'PyMC == 2.2', 'pandas >= 0.10.0', 'matplotlib >= 1.0.0'],
    setup_requires=['NumPy >= 1.6.0', 'PyMC == 2.2', 'pandas >= 0.10.0', 'matplotlib >= 1.0.0']
)
